@@ -17,7 +17,7 @@ coverity_scan() {
 	fi
 
 	export COVERITY_SCAN_PROJECT_NAME="${TRAVIS_REPO_SLUG}"
-	export COVERITY_SCAN_NOTIFICATION_EMAIL="vapier@gentoo.org"
+	export COVERITY_SCAN_NOTIFICATION_EMAIL="grobian@gentoo.org"
 	export COVERITY_SCAN_BUILD_COMMAND="make -j${ncpus}"
 	export COVERITY_SCAN_BUILD_COMMAND_PREPEND="git clean -q -x -d -f; git checkout -f"
 	export COVERITY_SCAN_BRANCH_PATTERN="master"
@@ -30,9 +30,24 @@ main() {
 	export CPPFLAGS="-I${PWD}/../sysroot"
 	export LDFLAGS="-L${PWD}/../sysroot"
 
-	# Standard optimized build.
-	m
-	m check
+	# ignore timestamps which git doesn't preserve
+	# disable openmp because Clang's libomp isn't installed
+	DEFARGS="--disable-maintainer-mode --disable-openmp"
+
+	do_run() {
+		v ./configure ${*}
+
+		# Standard optimized build.
+		m V=1
+		m check
+		m clean
+	}
+
+	do_run ${DEFARGS}
+	do_run ${DEFARGS} --enable-qmanifest --enable-qtegrity
+	do_run ${DEFARGS} --disable-qmanifest --enable-qtegrity
+	do_run ${DEFARGS} --enable-qmanifest --disable-qtegrity
+	do_run ${DEFARGS} --disable-qmanifest --disable-qtegrity
 
 	# LSan needs sudo, which we don't use at the moment
 	# Debug build w/ASAN and such enabled.
